/*
 * Copyright 2017-2019 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.StringEscapeUtils;
import org.diirt.datasource.ChannelHandler;
import org.diirt.datasource.DataSource;
import org.diirt.vtype.VByteArray;
import org.diirt.vtype.VString;
import org.diirt.vtype.VType;

import com.aquenos.epics.jackie.client.ChannelAccessClient;
import com.aquenos.epics.jackie.client.ChannelAccessClientConfiguration;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.common.util.ListenerLockPolicy;
import com.aquenos.epics.jackie.diirt.datasource.internal.JackieChannelHandler;

/**
 * <p>
 * Data source providing Channel Access PVs by using the EPICS Jackie library.
 * This data-source has type adapters for reading values using the usualy types
 * from the {@link VType} type hierarchy. When writing values, these VTypes are
 * also accepted, but the simple types ({@link String}s, boxed types of the
 * primitives or arrays of primitives) are preferred.
 * </p>
 * 
 * <p>
 * The channel names used when {@linkplain #createChannel(String) creating
 * channels} for this data source can be the pure name of the Channel Access
 * channel or it can be the name of the Channel Access channel with an
 * additional options string appended. If such an options string is present, it
 * must represent a valid JSON object and must follow the actual channel name,
 * separated by a single space character. At the moment, two different options
 * are supported:
 * </p>
 * 
 * <p>
 * The <em>longString</em> option specifies how a channel that has a native
 * data-type of <code>DBR_CHAR</code> should be treated. If <em>longString</em>
 * is <em>true</em>, the value is treated as a null-terminated string, resulting
 * in a {@link VString} being returned where otherwise a {@link VByteArray}
 * would be returned. If <em>false</em>, a <code>DBR_CHAR</code> is always
 * treated as a {@code VByteArray}.
 * </p>
 * 
 * <p>
 * If the <em>longString</em> option is not specified, it is automatically
 * determined based on the channel name. If the channel name ends with
 * ".FIELD$", where <em>FIELD</em> is any string that starts with an upper-case
 * letter and consists only of upper-case letters and digits,
 * <em>longString</em> is set to <em>true</em>. Otherwise, it is set to
 * <em>false</em>.
 * </p>
 * 
 * <p>
 * The <em>putCallback</em> option specifies whether a write operation succeeds
 * immediately, or whether the client waits for the server to acknowledge the
 * write operation. By setting <em>putCallback</em> to <em>true</em>, the code
 * creating the channel effectively tells this implementation that it does not
 * really care whether the write was actually successful. As there is a
 * difference in the protocol between a put with and without a callback and most
 * other Channel Access clients use the no-callback variant by default, the
 * <em>putCallback</em> option is set to <em>false</em> by default.
 * </p>
 * 
 * <p>
 * Examples of valid options strings are:
 * </p>
 * 
 * <ul>
 * <li>{"longString":"false","putCallback:true"}</li>
 * <li>{"longString":"true","putCallback:false"}</li>
 * <li>{"putCallback":"true"}</li>
 * <li>{}</li>
 * </ul>
 * 
 * @author Sebastian Marsching
 */
public class JackieDataSource extends DataSource {

    private ChannelAccessClient client;
    private volatile boolean closed;
    private JackieDataSourceConfiguration configuration;
    private final Logger logger = Logger
            .getLogger(getClass().getCanonicalName());

    /**
     * Creates a data source using the specified configuration.
     * 
     * @param configuration
     *            configuration that shall be used by this data source.
     */
    public JackieDataSource(JackieDataSourceConfiguration configuration) {
        super(true);
        ChannelAccessClientConfiguration clientConfiguration = configuration
                .toChannelAccessClientConfiguration();
        // We destroy channels and monitors while holding the same lock that we
        // take from the corresponding listeners, so we must not use the BLOCK
        // policy. The IGNORE policy would work for us, because we handle
        // notifications after destruction correctly, but it is still good to be
        // reminded of this by having to catch the corresponding exceptions
        // explicitly, so we prefer REPORT over IGNORE.
        this.client = new DefaultChannelAccessClient(clientConfiguration,
                ListenerLockPolicy.REPORT);
        this.configuration = configuration;
        logger.fine("Data source initialized.");
    }

    @Override
    protected ChannelHandler createChannel(String channelName) {
        if (closed) {
            throw new IllegalStateException(
                    "Cannot create a channel after this data-source has been closed.");
        }
        // When debugging, it might be useful to know which data-source was used
        // to create a certain channel.
        logger.fine("Creating channel \""
                + StringEscapeUtils.escapeJava(channelName) + "\".");
        try {
            return new JackieChannelHandler(client, channelName, configuration,
                    logger);
        } catch (RuntimeException e) {
            // We log the exception with a level of FINE because it should
            // usually be treated at a higher level. The logging here is only
            // intended for when the code calling this method somehow discards
            // the exception without logging it (or does not provide the
            // complete stack trace).
            logger.log(Level.FINE,
                    "Error while trying to create channel \""
                            + StringEscapeUtils.escapeJava(channelName) + "\": "
                            + e.getMessage(),
                    e);
            throw e;
        }
    }

    @Override
    public void close() {
        // Obviously, there is a race condition here, so this method might run
        // more than once. However, this does not matter because
        // client.destroy() can be called multiple times and calling the super
        // class's close() method multiple times should not matter either
        // because if it did, it should take care of this itself.
        // For this reason, this check is simply to avoid running code more
        // often than needed. We need the closed flag anyway in order to throw
        // an IllegalStateException in createChannel(...).
        if (closed) {
            return;
        }
        closed = true;
        try {
            super.close();
        } finally {
            client.destroy();
        }
        logger.fine("Data source closed.");
    }

}
