package com.aquenos.epics.jackie.diirt.vtype;

import java.nio.ShortBuffer;
import java.time.Instant;
import java.util.ArrayList;
import java.util.List;

import org.diirt.util.array.ArrayInt;
import org.diirt.util.array.ListInt;
import org.diirt.vtype.VEnumArray;
import org.diirt.vtype.VTypeToString;

import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeEnum;

/**
 * {@link VEnumArray} implementation that is backed by a
 * {@link ChannelAccessGraphicsEnum} and {@link ChannelAccessTimeEnum}.
 * 
 * @author Sebastian Marsching
 */
public final class JackieVEnumArray extends JackieEnum implements VEnumArray {

    /**
     * <p>
     * Creates a value backed by the specified {@code graphicsValue} and
     * {@code timeValue}. The {@code graphicsValue} is used for the meta-data
     * (labels) and the {@code timeValue} is used for the alarm status,
     * time-stamp, and value.
     * </p>
     * 
     * <p>
     * If {@code disconnectTime} is not <code>null</code>, it is used as the
     * time-stamp and the alarm severity is set to {@code UNDEFINED}.
     * </p>
     * 
     * @param graphicsValue
     *            value used for meta-data (labels).
     * @param timeValue
     *            value used for alarm status, time-stamp, and value.
     * @param disconnectTime
     *            time at which the channel was disconnected. Usually this
     *            should be <code>null</code>. If it is not <code>null</code>,
     *            the alarm severity is set to {@code UNDEFINED} and this
     *            parameter is used as the time-stamp (instead of using the
     *            time-stamp from the {@code timeValue}.
     */
    public JackieVEnumArray(ChannelAccessGraphicsEnum graphicsValue,
            ChannelAccessTimeEnum timeValue, Instant disconnectTime) {
        super(graphicsValue, timeValue, disconnectTime, false);
    }

    @Override
    public ListInt getSizes() {
        return new ArrayInt(timeValue.getValueSize());
    }

    @Override
    public List<String> getData() {
        List<String> labels = getLabels();
        ShortBuffer indices = timeValue.getValue();
        ArrayList<String> values = new ArrayList<String>(indices.remaining());
        while (indices.hasRemaining()) {
            int index = indices.get() & 0xffff;
            if (index < labels.size()) {
                values.add(labels.get(index));
            } else {
                values.add(Integer.toString(index));
            }
        }
        return values;
    }

    @Override
    public ListInt getIndexes() {
        ShortBuffer buffer = timeValue.getValue();
        return new ListInt() {
            @Override
            public int getInt(int index) {
                // Enums are unsigned 16-bit integers, so we have to make
                // sure they are treated as unsigned.
                return buffer.get(index) & 0xffff;
            }

            @Override
            public int size() {
                return buffer.remaining();
            }
        };
    }

    @Override
    public String toString() {
        return VTypeToString.toString(this);
    }

}
